package re.jimmitry.apps974.furious;
import re.jimmitry.apps974.furious.*;
import re.jimmitry.apps974.fursys.*;
import re.jimmitry.apps974.wsfurious.*;
import re.jimmitry.apps974.wsfurious.xsd.*;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.KeyStore;
import java.security.Provider;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.Signature;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateCrtKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.RSAPrivateCrtKeySpec;
import java.security.spec.RSAPublicKeySpec;
import java.security.Security;
import java.util.Calendar;
import java.util.Date;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import org.apache.commons.codec.binary.Hex;
import org.apache.log4j.Logger;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;
import org.bouncycastle.asn1.ASN1Encodable;
import org.bouncycastle.asn1.ASN1EncodableVector;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.DERBMPString;
import org.bouncycastle.asn1.DERBitString;
import org.bouncycastle.asn1.DERInteger;
import org.bouncycastle.asn1.DERNull;
import org.bouncycastle.asn1.DERObjectIdentifier;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.DEROutputStream;
import org.bouncycastle.asn1.DERSequence;
import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
import org.bouncycastle.asn1.x509.BasicConstraints;
import org.bouncycastle.asn1.x509.DigestInfo;
import org.bouncycastle.asn1.x509.RSAPublicKeyStructure;
import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;
import org.bouncycastle.asn1.x509.TBSCertificateStructure;
import org.bouncycastle.asn1.x509.Time;
import org.bouncycastle.asn1.x509.V3TBSCertificateGenerator;
import org.bouncycastle.asn1.x509.X509CertificateStructure;
import org.bouncycastle.asn1.x509.X509Extension;
import org.bouncycastle.asn1.x509.X509Extensions;
import org.bouncycastle.asn1.x509.X509Name;
import org.bouncycastle.asn1.x509.X509ObjectIdentifiers;
import org.bouncycastle.crypto.AsymmetricBlockCipher;
import org.bouncycastle.crypto.AsymmetricCipherKeyPair;
import org.bouncycastle.crypto.CryptoException;
import org.bouncycastle.crypto.DataLengthException;
import org.bouncycastle.crypto.digests.SHA1Digest;
import org.bouncycastle.crypto.encodings.PKCS1Encoding;
import org.bouncycastle.crypto.engines.RSAEngine;
import org.bouncycastle.crypto.generators.RSAKeyPairGenerator;
import org.bouncycastle.crypto.params.RSAKeyGenerationParameters;
import org.bouncycastle.crypto.params.RSAKeyParameters;
import org.bouncycastle.crypto.params.RSAPrivateCrtKeyParameters;
import org.bouncycastle.jce.PrincipalUtil;
import org.bouncycastle.jce.interfaces.PKCS12BagAttributeCarrier;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.jce.provider.JDKPKCS12KeyStore;
import org.bouncycastle.jce.provider.X509CertificateObject;
//import org.bouncycastle.x509.X509Util;
import org.bouncycastle.x509.extension.AuthorityKeyIdentifierStructure;
import org.bouncycastle.x509.extension.SubjectKeyIdentifierStructure;


/** This class uses the Bouncycastle lightweight API to generate X.509 certificates programmatically.
 * It assumes a CA certificate and its private key to be available and can sign the new certificate with
 * this CA. Some of the code for this class was taken from 
 * org.bouncycastle.x509.X509V3CertificateGenerator, but adapted to work with the lightweight API instead of
 * JCE (which is usually not available on MIDP2.0). 
 * 
 * @author Rene Mayrhofer
 * @version 1.0
 */
public class X509CertificateGenerator {
	/** Our log4j logger. */
	private static Logger logger = Logger.getLogger(X509CertificateGenerator.class);
	
	/** This method is used for signing the certificate. */
	public static final String CertificateSignatureAlgorithm = "SHA1WithRSAEncryption";
	/** This string is used as the friendly name for the certificate in the PKCS12 exported file. */
	public static final String CertificateExportFriendlyName = "Certificate for IPSec WLAN access";
	/** This string is used as the friendly name for the private key in the PKCS12 exported file. */
	public static final String KeyExportFriendlyName = "Private key for IPSec WLAN access"; 
	
	/** This holds the certificate of the CA used to sign the new certificate. The object is created in the constructor. */
	private X509Certificate caCert;
	/** This holds the private key of the CA used to sign the new certificate. The object is created in the constructor. */
	private RSAPrivateCrtKeyParameters caPrivateKey;
	
	/** Creates a new self-signed CA (certificate authority) for subsequently signing certificates. 
	 * 
	 * @param commonName The common name (CN) field of the X.509 distinguished name that should be set
	 *                   for the new certificate. All other fields of the distinguished name are not set.
	 * @param validityDays How long the new certificate should be valid, in days.
	 * @param caFile The PKCS12 encoded file to which the CA should be exported to. It will contain the
	 *               self-signed certificate and the matching private key.
	 * @param caPassword The password used to encode the PKCS12 file.
	 * @param caAlias The friendly name of the CA in the PKCS12 file.
	 * 
	 *                 operations. If set to false, the JCE infrastructure with the configured default provider
	 *                 will be used. JCE may be faster depending on the provider implementation, but it might
	 *                 not be available on embedded platforms, i.e. J2ME.
	 * @return true if the CA could be created, self-signed, and exported successfully, false otherwise.
	 * @throws CryptoException 
	 * @throws IOException 
	 * @throws InvalidKeySpecException 
	 * @throws CertificateException 
	 * @throws NoSuchProviderException 
	 * @throws KeyStoreException 
	 * @throws NoSuchAlgorithmException 
	 * @throws SignatureException 
	 * @throws SecurityException 
	 * @throws DataLengthException 
	 * @throws InvalidKeyException 
	 */
public static Date getNotAfterDay(String caFile, String caPassword, String caAlias) 
	throws NoSuchAlgorithmException, CertificateException, FileNotFoundException, IOException, InvalidKeyException, NoSuchProviderException, SignatureException {

		if (caFile == null || caPassword == null || caAlias == null) {
			throw new IllegalArgumentException("Can not work with null parameter");
		}
		
		
		PKCS12Content content = loadFromKeyStore(new FileInputStream(new File(caFile)), caPassword, caAlias);
		if (content.privateKeys == null || content.privateKeys[0] == null) {
			throw new RuntimeException("Got null key from keystore!"); 
		}
		if (content.certificates == null || content.certificates[0] == null) {
			throw new RuntimeException("Got null cert from keystore!"); 
		}
		
		return content.certificates[0].getNotAfter();
	}

public static BigInteger getSerialNumber(String caFile, String caPassword, String caAlias) 
	throws NoSuchAlgorithmException, CertificateException, FileNotFoundException, IOException, InvalidKeyException, NoSuchProviderException, SignatureException {

		if (caFile == null || caPassword == null || caAlias == null) {
			throw new IllegalArgumentException("Can not work with null parameter");
		}
		
		
		PKCS12Content content = loadFromKeyStore(new FileInputStream(new File(caFile)), caPassword, caAlias);
		if (content.privateKeys == null || content.privateKeys[0] == null) {
			throw new RuntimeException("Got null key from keystore!"); 
		}
		if (content.certificates == null || content.certificates[0] == null) {
			throw new RuntimeException("Got null cert from keystore!"); 
		}
		
		return content.certificates[0].getSerialNumber();
	}
	public static boolean createNewCa(String commonName, int validityDays, 
			String caFile, String caPassword, String caAlias) throws InvalidKeyException, DataLengthException, SecurityException, SignatureException, NoSuchAlgorithmException, KeyStoreException, CertificateException, InvalidKeySpecException, IOException, CryptoException {
		X509CertificateGenerator g = new X509CertificateGenerator();
		return g.createCertificate(commonName, validityDays, caFile, caPassword, caAlias);
	}

	/** Converts from a PKCS12 encoded file to PEM encoded files readable by openssl (and subsequently 
	 * e.g. openswan and racoon).
	 * 
	 * @param inFile The PKCS12 encoded input file. It is assumed to contain the (client) certificate,
	 *               the matching private key, and the complete certificate chain up to the self-signed
	 *               root CA.
	 * @param inPassword The password needed to decrypt the PKCS12 encoded input file.
	 * @param outCertFile The output certificate in PEM format. If null, it will not be created.
	 * @param outKeyFile The output private key in PEM format. If null, it will not be created.
	 * @param outCertChainFile The output certificate chain in PEM format. If null, it will not be created.
	 * 
	 *                 operations. If set to false, the JCE infrastructure with the configured default provider
	 *                 will be used. JCE may be faster depending on the provider implementation, but it might
	 *                 not be available on embedded platforms, i.e. J2ME.
	 * @return true if all requested parts could be exported successfully, false otherwise.
	 */
	public static boolean convertPKCS12toPEM(String inFile, String inPassword, 
			String outCertFile, String outKeyFile, String outCertChainFile) {
		
		if (inFile == null || inPassword == null || 
				outCertFile == null || outKeyFile == null || outCertChainFile == null) {
			throw new IllegalArgumentException("Can not work with null parameter");
		}
		
		// echo "test" | openssl pkcs12 -in test.p12 -passin stdin -nodes -clcerts -nokeys > /tmp/gateway.pem
		// echo "test" | openssl pkcs12 -in test.p12 -passin stdin -nodes -cacerts -nokeys > /tmp/ipsecca.pem
		// echo "test" | openssl pkcs12 -in test.p12 -passin stdin -nodes -nocerts > /tmp/gatewayKey.pem
		
		

		return false;
	}

	/** Initializes the objects without an existing CA. This is useful to create a new CA, because
	 * createCertificate will create a self-signed certificate if no CA has been set. This 
	 * constructor is used by createNewCa.
	 * 
	 * @see #createNewCa
	 * @see #createCertificate(String, int, String, String, String)
	 * 
	 *                 operations. If set to false, the JCE infrastructure with the configured default provider
	 *                 will be used. JCE may be faster depending on the provider implementation, but it might
	 *                 not be available on embedded platforms, i.e. J2ME.
	 */
	protected X509CertificateGenerator() {
		caPrivateKey = null;
		caCert = null;
	}
	
	/** Initializes the object for creating certificates by loading the CA certificate and private key. 
	 * 
	 * A new CA can be created with:
	 * 
	 * Comment out basicConstraints in /etc/ssl/openssl.cnf (CA:FALSE should not be set, but it does not need to be set to true)
	 * /usr/lib/ssl/misc/CA.sh -newca
	 * openssl pkcs12 -export -in demoCA/cacert.pem -inkey demoCA/private/cakey.pem -out ca.p12 -name "Test CA"
	 *
	 * @param caFile The PKCS12 encoded file containing the whole CA to use. It must contain the CA certificate
	 *               (which will be self-signed for top-level CAs) and the matching private key.
	 * @param caPassword The password necessary to decode the PKCS12 file.
	 * @param caAlias The friendly name of the CA in the PKCS12 file.
	 * 
	 *                 operations. If set to false, the JCE infrastructure with the configured default provider
	 *                 will be used. JCE may be faster depending on the provider implementation, but it might
	 *                 not be available on embedded platforms, i.e. J2ME.
	 */
	public X509CertificateGenerator(String caFile, String caPassword, String caAlias) 
			throws NoSuchAlgorithmException, CertificateException, FileNotFoundException, IOException, InvalidKeyException, NoSuchProviderException, SignatureException {
		
		if (caFile == null || caPassword == null || caAlias == null) {
			throw new IllegalArgumentException("Can not work with null parameter");
		}
		
		
		PKCS12Content content = loadFromKeyStore(new FileInputStream(new File(caFile)), caPassword, caAlias);
		if (content.privateKeys == null || content.privateKeys[0] == null) {
			throw new RuntimeException("Got null key from keystore!"); 
		}
		if (content.certificates == null || content.certificates[0] == null) {
			throw new RuntimeException("Got null cert from keystore!"); 
		}

		caPrivateKey = content.privateKeys[0];
		caCert = content.certificates[0];

		caCert.verify(caCert.getPublicKey());
		
	}
	
	/** Returns the distinguished name of the CA used to sign the newly generated certificates.
	 * This is important e.g. when using the new certificates for IPSec connections under
	 * Windows 2000/XP, because the CA DN must be specified for such connections.
	 * @return The CA DN.
	 */
	public String getCaDistinguishedName() {
		
		if (caCert == null) {
			return null;
		}
		return caCert.getSubjectDN().toString();
	}
	
	/** This method should create something similar to:
	 *
	 * openssl req -new -outform PEM -newkey rsa:1024 -nodes -keyout /tmp/test.key -keyform PEM -out /tmp/test.pem -days 30 -config /etc/ssl/openssl.cnf
	 * openssl ca -policy policy_anything -out /tmp/test.crt -config /etc/ssl/openssl.cnf -infiles  /tmp/test.pem
	 * openssl pkcs12 -export -in /tmp/test.crt -inkey /tmp/test.key -certfile demoCA/cacert.pem -out test.p12
	 * 
	 * @param commonName The common name (CN) field of the X.509 distinguished name that should be set
	 *                   for the new certificate. All other fields of the distinguished name are not set.
	 * @param validityDays How long the new certificate should be valid, in days.
	 * @param exportFile The PKCS12 encoded file to which the certificate should be exported to. It will contain the
	 *               self-signed certificate and the matching private key.
	 * @param exportPassword The password used to encode the PKCS12 file.
	 * @return true if the certificate could be created, signed, and exported successfully, false otherwise.
	 */
	public boolean createCertificate(String commonName, int validityDays, String exportFile, String exportPassword) throws 
			IOException, InvalidKeyException, SecurityException, SignatureException, NoSuchAlgorithmException, DataLengthException, CryptoException, KeyStoreException, CertificateException, InvalidKeySpecException {
		return createCertificate(commonName, validityDays, exportFile, exportPassword, null);
	}
	
	/** This method implements the public one, but offers an additional parameter which is only used when
	 * creating a new CA, namely the export alias to use.
	 * @param commonName @see #createCertificate(String, int, String, String)
	 * @param validityDays @see #createCertificate(String, int, String, String)
	 * @param exportFile @see #createCertificate(String, int, String, String)
	 * @param exportPassword @see #createCertificate(String, int, String, String)
	 * @param exportAlias If this additional parameter is null, a default value will be used as the "friendly name" in the PKCS12 file.
	 * @return @see #createCertificate(String, int, String, String)
	 * 
	 * @see #X509CertificateGenerator(boolean)
	 */
	protected boolean createCertificate(String commonName, int validityDays, String exportFile, String exportPassword, String exportAlias) throws 
			IOException, InvalidKeyException, SecurityException, SignatureException, NoSuchAlgorithmException, DataLengthException, CryptoException, KeyStoreException, CertificateException, InvalidKeySpecException {
			//javax.swing.JOptionPane.showMessageDialog(null,"C0", "974 Application Server", javax.swing.JOptionPane.INFORMATION_MESSAGE);
		SecureRandom sr = new SecureRandom();
		//SecureRandom sr = SecureRandom.getInstance("SHA1PRNG");
		//javax.swing.JOptionPane.showMessageDialog(null,"C0"+sr.getProvider(), "974 Application Server", javax.swing.JOptionPane.INFORMATION_MESSAGE);
		// the JCE representation
		PublicKey pubKey;
		PrivateKey privKey;
		// the BCAPI representation
		RSAPrivateCrtKeyParameters privateKey = null;
		// generate the keypair for the new certificate
			// this is the JSSE way of key generation
			KeyPairGenerator keyGen = KeyPairGenerator.getInstance("RSA");
			keyGen.initialize(1024, sr);
			KeyPair keypair = keyGen.generateKeyPair();
			privKey = keypair.getPrivate();
			pubKey = keypair.getPublic();
		
		Calendar expiry = Calendar.getInstance();
		expiry.add(Calendar.DAY_OF_YEAR, validityDays);

		X509Name x509Name = new X509Name(commonName);
		
		V3TBSCertificateGenerator certGen = new V3TBSCertificateGenerator();
		  certGen.setSerialNumber(new DERInteger(BigInteger.valueOf(System.currentTimeMillis())));
	    if (caCert != null) {
	    	// Attention: this is a catch! Just using "new X509Name(caCert.getSubjectDN().getName())" will not work!
	    	// I don't know why, because the issuerDN strings look similar with both versions.
	    	certGen.setIssuer(PrincipalUtil.getSubjectX509Principal(caCert));
	    	
	    }
	    else {
	    	// aha, no CA set, which means that we should create a self-signed certificate (called from createCA)
	    	certGen.setIssuer(x509Name);
	    	
	    }
		certGen.setSubject(x509Name);
		DERObjectIdentifier sigOID = X509Util.getAlgorithmOID(CertificateSignatureAlgorithm);
		AlgorithmIdentifier sigAlgId = new AlgorithmIdentifier(sigOID, new DERNull());
		certGen.setSignature(sigAlgId);
		//certGen.setSubjectPublicKeyInfo(new SubjectPublicKeyInfo(sigAlgId, pkStruct.toASN1Object()));
		// TODO: why does the coding above not work? - make me work without PublicKey class
		certGen.setSubjectPublicKeyInfo(new SubjectPublicKeyInfo((ASN1Sequence)new ASN1InputStream(
                new ByteArrayInputStream(pubKey.getEncoded())).readObject()));
		certGen.setStartDate(new Time(new Date(System.currentTimeMillis())));
		certGen.setEndDate(new Time(expiry.getTime()));
		// These X509v3 extensions are not strictly necessary, but be nice and provide them...
	    Hashtable extensions = new Hashtable();
	    Vector extOrdering = new Vector();
	    addExtensionHelper(X509Extensions.SubjectKeyIdentifier, false, new SubjectKeyIdentifierStructure(pubKey), extOrdering, extensions);
		if (caCert != null) {
			// again: only if we have set CA
			addExtensionHelper(X509Extensions.AuthorityKeyIdentifier, false, new AuthorityKeyIdentifierStructure(caCert), extOrdering, extensions);
		} 
		else {
			// but if we create a new self-signed cert, set its capability to be a CA
			// this is a critical extension (true)!
			addExtensionHelper(X509Extensions.BasicConstraints, true, new BasicConstraints(0), extOrdering, extensions);
		}
		String dnsNameList = "", dnsName="";
		try {
		  dnsNameList = Fur_File.getINIinfos("PARAM.INI", "CERTIFICATION", "MULTI_HOST");;
    } catch (Exception e) {}
    ASN1EncodableVector sanVector = new ASN1EncodableVector();
    for(int y=1;y<=Integer.parseInt(Fur_Atome.getSizeAtomeList(dnsNameList, ","));y++) {
      dnsName=Fur_Atome.getAtome(dnsNameList, ""+y, ",");
      sanVector.add( new GeneralName( GeneralName.dNSName, Fur_String.LRTRIM(dnsName) ) );
    }
    DERSequence san = new DERSequence( sanVector );

    GeneralNames sanGeneralNames = new GeneralNames( san );
    //certGen.addExtension( X509Extensions.SubjectAlternativeName, true, sanGeneralNames );

		extOrdering.addElement(X509Extensions.SubjectAlternativeName);
		extensions.put(X509Extensions.SubjectAlternativeName, new X509Extension(true, new DEROctetString(sanGeneralNames)));
		certGen.setExtensions(new X509Extensions(extOrdering, extensions));
		// attention: hard coded to be SHA1+RSA!
		SHA1Digest digester = new SHA1Digest();
		AsymmetricBlockCipher rsa = new PKCS1Encoding(new RSAEngine());
		TBSCertificateStructure tbsCert = certGen.generateTBSCertificate();
		ByteArrayOutputStream bOut = new ByteArrayOutputStream();
		DEROutputStream dOut = new DEROutputStream(bOut);
		dOut.writeObject(tbsCert);
		// and now sign
		byte[] signature;
		
			// or the JCE way
	        Signature sig = Signature.getInstance(sigOID.getId());
			if (caCert != null) {
		        PrivateKey caPrivKey = KeyFactory.getInstance("RSA").generatePrivate(
		        		new RSAPrivateCrtKeySpec(caPrivateKey.getModulus(), caPrivateKey.getPublicExponent(),
		        				caPrivateKey.getExponent(), caPrivateKey.getP(), caPrivateKey.getQ(), 
		        				caPrivateKey.getDP(), caPrivateKey.getDQ(), caPrivateKey.getQInv()));
				sig.initSign(caPrivKey, sr);
			} else {
				
				sig.initSign(privKey, sr);
			}
	        sig.update(bOut.toByteArray());
	        signature = sig.sign();
		// and finally construct the certificate structure
        ASN1EncodableVector  v = new ASN1EncodableVector();

        v.add(tbsCert);
        v.add(sigAlgId);
        v.add(new DERBitString(signature));
        X509CertificateObject clientCert = new X509CertificateObject(new X509CertificateStructure(new DERSequence(v))); 
       /*if (caCert != null) {
        	clientCert.verify(caCert.getPublicKey());
        }
        else {
        	clientCert.verify(pubKey);
        }*/

        // and export as PKCS12 formatted file along with the private key and the CA certificate 
        PKCS12BagAttributeCarrier bagCert = clientCert;
        // if exportAlias is set, use that, otherwise a default name
        bagCert.setBagAttribute(PKCSObjectIdentifiers.pkcs_9_at_friendlyName,
        		new DERBMPString(exportAlias == null ? CertificateExportFriendlyName : exportAlias));
        bagCert.setBagAttribute(
                PKCSObjectIdentifiers.pkcs_9_at_localKeyId,
                new SubjectKeyIdentifierStructure(pubKey));
        // this does not work as in the example
        /*PKCS12BagAttributeCarrier   bagKey = (PKCS12BagAttributeCarrier)privKey;
        bagKey.setBagAttribute(
            PKCSObjectIdentifiers.pkcs_9_at_localKeyId,
            new SubjectKeyIdentifierStructure(tmpKey));*/
        Object store= null;

		Security.addProvider(new BouncyCastleProvider());
		try {
			store = java.security.KeyStore.getInstance("PKCS12", "BC");
			} catch (Exception rt) {}
			((java.security.KeyStore) store).load(null, null);

        FileOutputStream fOut = new FileOutputStream(exportFile);
        X509Certificate[] chain;
        
        if (caCert != null) {
        	chain = new X509Certificate[2];
            // first the client, then the CA certificate - this is the expected order for a certificate chain
            chain[0] = clientCert;
            chain[1] = caCert;
        }
        else {
        	// for a self-signed certificate, there is no chain...
        	chain = new X509Certificate[1];
        	chain[0] = clientCert;
        }
        //javax.swing.JOptionPane.showMessageDialog(null,"Setp 1/2: KeyEntry > GO ...", "974 Application Server", javax.swing.JOptionPane.INFORMATION_MESSAGE);
        		((java.security.KeyStore) store).setKeyEntry(exportAlias == null ? KeyExportFriendlyName : exportAlias, privKey, exportPassword.toCharArray(), chain);
        //javax.swing.JOptionPane.showMessageDialog(null,"Setp 1/2: KeyEntry > OK", "974 Application Server", javax.swing.JOptionPane.INFORMATION_MESSAGE);
        		//javax.swing.JOptionPane.showMessageDialog(null,"Setp 2/2: Storing > GO ...", "974 Application Server", javax.swing.JOptionPane.INFORMATION_MESSAGE);
        		((java.security.KeyStore) store).store(fOut, exportPassword.toCharArray());
        //javax.swing.JOptionPane.showMessageDialog(null,"Setp 2/2: Storing > OK", "974 Application Server", javax.swing.JOptionPane.INFORMATION_MESSAGE);
	fOut.close();
        return true;
	}
	
	/** This is only a small helper function for adding X.509v3 extensions 
	 * @throws IOException */
	private void addExtensionHelper(DERObjectIdentifier extId, boolean critical, ASN1Encodable extVal, Vector extensionsOrder, Hashtable extensions) throws IOException {
		
		ByteArrayOutputStream bOut = new ByteArrayOutputStream();
		DEROutputStream dOut = new DEROutputStream(bOut);
		dOut.writeObject(extVal);
		extensions.put(extId, new X509Extension(critical, new DEROctetString(bOut.toByteArray())));
		extensionsOrder.addElement(extId);
    }
	
	private static class PKCS12Content {
		X509Certificate[] certificates;
		String[] certificateAliases;
		RSAPrivateCrtKeyParameters[] privateKeys;
		String[] privateKeyAliases;
	}
	
	public static boolean writeCERFile(InputStream keystore, String password, String aliasStartsWith, String CerFile) {
	  boolean b =true;
	  PKCS12Content ret = new PKCS12Content();
		try {
			Object caKs;
			String[] aliases;
			X509Certificate[] certificates;
			Key[] keys;
			int numKeys = 0, numCertificates = 0, numEntries;
				caKs = java.security.KeyStore.getInstance("PKCS12");
				((java.security.KeyStore) caKs).load(keystore, password.toCharArray());
				
				// extract all known aliases and remember if they are keys or certificates
				Enumeration entries = ((java.security.KeyStore) caKs).aliases();
				numEntries = ((java.security.KeyStore) caKs).size();
				aliases = new String[numEntries];
				certificates = new X509Certificate[numEntries];
				keys = new Key[numEntries];
				for (int i=0; i<numEntries; i++) {
					aliases[i] = (String) entries.nextElement();
					//Fur_File.Trace("FOR"+i+"/"+numEntries+":"+aliases[i]);
					//if (((java.security.KeyStore) caKs).isCertificateEntry(aliases[i])) {
					  certificates[i] = (X509Certificate) ((java.security.KeyStore) caKs).getCertificate(aliases[i]);//Fur_File.Trace("CER"+i+"/"+numEntries+":"+aliases[i]);
					
						FileOutputStream fos = new FileOutputStream(CerFile);
            fos.write(certificates[i].getEncoded());
            fos.flush();
            fos.close();
            /*} else if (((java.security.KeyStore) caKs).isKeyEntry(aliases[i])) {Fur_File.Trace("KEY"+i+"/"+numEntries+":"+aliases[i]);} 
            else Fur_File.Trace("OTHER"+i+"/"+numEntries+":"+aliases[i]);*/
				}
			
		} 
		catch (Exception e) {
			Fur_File.Trace("CER GENERATING ERROR: "+e.getMessage());b=false;
		}
		return b;
	}
	
	/** This is a helper function for loading from a PKCS12 file.
	 * @param keystore The keystore/PKCS12 file to load from.
	 * @param password The password used to encrypt the keystore/PKCS12 file.
	 * @param aliasStartsWith If not null, then only those elements of the file/keystore
	 *                        whose alias starts with this string will be returned. 
	 * 
	 * @return The elements if they could be successfully loaded, null otherwise.
	 */
	private static PKCS12Content loadFromKeyStore(InputStream keystore, 
			String password, String aliasStartsWith) {
		
		PKCS12Content ret = new PKCS12Content();
		
		try {
			Object caKs;
			String[] aliases;
			X509Certificate[] certificates;
			Key[] keys;
			int numKeys = 0, numCertificates = 0, numEntries;
			
			
				caKs = java.security.KeyStore.getInstance("PKCS12");
				((java.security.KeyStore) caKs).load(keystore, password.toCharArray());
				
				// extract all known aliases and remember if they are keys or certificates
				Enumeration entries = ((java.security.KeyStore) caKs).aliases();
				numEntries = ((java.security.KeyStore) caKs).size();
				aliases = new String[numEntries];
				certificates = new X509Certificate[numEntries];
				keys = new Key[numEntries];
				for (int i=0; i<numEntries; i++) {
					aliases[i] = (String) entries.nextElement();
					if (aliasStartsWith != null && !aliases[i].startsWith(aliasStartsWith)) { 
						continue;
					}
						
					if (((java.security.KeyStore) caKs).isCertificateEntry(aliases[i])) {
						certificates[i] = (X509Certificate) ((java.security.KeyStore) caKs).getCertificate(aliases[i]);
						numCertificates++;
					}
					else if (((java.security.KeyStore) caKs).isKeyEntry(aliases[i])) {
						keys[i] = ((java.security.KeyStore) caKs).getKey(aliases[i], password.toCharArray()); 
						numKeys++;
						// this is a dirty hack - when an entry is marked as a key, it may also have a certificate with the same alias
						certificates[i] = (X509Certificate) ((java.security.KeyStore) caKs).getCertificate(aliases[i]);
						if (certificates[i] != null) {
							numCertificates++;
						}
					}
					
				}
			
			
			
			// now that we know how many elements we have, initialize
			ret.certificates = new X509Certificate[numCertificates];
			ret.privateKeys = new RSAPrivateCrtKeyParameters[numKeys];
			ret.certificateAliases = new String[numCertificates];
			ret.privateKeyAliases = new String[numKeys];
			// and copy
			int outCertInd = 0;
			for (int i=0; i<numEntries; i++) {
				if (certificates[i] != null) {
					ret.certificates[outCertInd] = certificates[i];
					ret.certificateAliases[outCertInd++] = aliases[i];
				}
			}
			int outKeyInd = 0;
			for (int i=0; i<numCertificates; i++) {
				if (keys[i] != null) {
					RSAPrivateCrtKey privKey = (RSAPrivateCrtKey) keys[i];
					ret.privateKeys[outKeyInd] = new RSAPrivateCrtKeyParameters(privKey.getModulus(), privKey.getPublicExponent(), privKey.getPrivateExponent(),
							privKey.getPrimeP(), privKey.getPrimeQ(), privKey.getPrimeExponentP(), privKey.getPrimeExponentQ(), privKey.getCrtCoefficient());
					ret.privateKeyAliases[outKeyInd++] = aliases[i];
				}
			}
		
			return ret;
		} 
		catch (IOException e) {
			return null;
		} catch (KeyStoreException e) {
			return null;
		} catch (NoSuchAlgorithmException e) {
			return null;
		} catch (CertificateException e) {
			return null;
		} catch (UnrecoverableKeyException e) {
			return null;
		}
	}
	
	/** This is a helper function for fetching the distinguished name from a certificate.
	 * 
	 * @param keystore The keystore/PKCS12 file to load from.
	 * @param password The password used to encrypt the keystore/PKCS12 file.
	 * @param alias The alias used to store the certificate in the keystore/PKCS12 file.
	 * 
	 * @return The subject DN if the certificate could be successfully loaded, null otherwise.
	 */
	public static String getCertificateDistinguishedName(InputStream keystore,
			String password, String alias) {
		
		PKCS12Content content = loadFromKeyStore(keystore, password, alias);
		if (content == null || content.certificates == null || content.certificates[0] == null)
			return null;
		return content.certificates[0].getSubjectDN().toString();
	}

	/** This is a helper function for fetching the validity from a certificate.
	 * 
	 * @param keystore The keystore/PKCS12 file to load from.
	 * @param password The password used to encrypt the keystore/PKCS12 file.
	 * @param alias The alias used to store the certificate in the keystore/PKCS12 file.
	 * 
	 * @return The validity in days (from now) if the certificate could be loaded successfully, 
	 *         -1 otherwise
	 */
	public static int getCertificateValidity(InputStream keystore,
			String password, String alias) {
		
		PKCS12Content content = loadFromKeyStore(keystore, password, alias);
		if (content == null || content.certificates == null || content.certificates[0] == null)
			return -1;
		return (int) ((content.certificates[0].getNotAfter().getTime() - System.currentTimeMillis()) / 1000 / 3600 / 24);
	}
	

	
	/** The test CA can e.g. be created with
	 * 
	 * Hmm, this CA doesn't work - look at the Javadoc comment for the constructor for how to create it correctly.
	 * echo -e "AT\nUpper Austria\nSteyr\nMy Organization\nNetwork tests\nTest CA certificate\nme@myserver.com\n\n\n" | \
	     openssl req -new -x509 -outform PEM -newkey rsa:2048 -nodes -keyout /tmp/ca.key -keyform PEM -out /tmp/ca.crt -days 365;
	   echo "test password" | openssl pkcs12 -export -in /tmp/ca.crt -inkey /tmp/ca.key -out ca.p12 -name "Test CA" -passout stdin
	 * 
	 * The created certificate can be displayed with
	 * 
	 * openssl pkcs12 -nodes -info -in test.p12 > /tmp/test.cert && openssl x509 -noout -text -in /tmp/test.cert
	 * 
	 * @param args
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		System.out.println(X509CertificateGenerator.createNewCa("EmailAddress=root@dbsconnector.re,CN=DBSCONNECTOR ERP Digital Certification,OU=SERVER DBSCONNECTOR ERP,O=DBSCONNECTOR Corporation,L=Ile de la Reunion,ST=France,C=RE", 365, "ca.p12", "test password", "dbsconnector_ca"));
		System.out.println(new X509CertificateGenerator("ca.p12", "test password", "dbsconnector_ca").createCertificate("EmailAddress=jimmitry@dbsconnector.re,CN=Mr Jimmitry PAYET (ADMIN),OU=CLIENT DBSCONNECTOR ERP,O=ADMIN/Administration;Service/DBSCONNECTOR,L=Ile de la Reunion,ST=France,C=RE", 180, "test.p12", "test", "dbsconnector_cu"));
		
	}
public static long getExpiration() {
  long b = 0;
  try {
  b =(((X509CertificateGenerator.getNotAfterDay("webapps/DBSConnector/SECURITY/PKCS12/DBSCONNECTOR/USERS/dbsconnector/PKS/dbsconnector.p12", FurPARAM.getDBSConnectorParam("CU_PASSWORD").DATA, "dbsconnector_cu")).getTime())-((new Date()).getTime()))/1000/3600/24;
  } catch (Exception e) {}
  return b;
}
}
